@@ -4,7 +4,7 @@
  *   Copyright (C) 2007-2009 Richard Goedeken                              *
  *   Copyright (C) 2007-2008 Ebenblues                                     *
  *   Copyright (C) 2003 JttL                                               *
- *   Copyright (C) 2002 Hactarux                                           *
+ *   Copyright (C) 2002 Hacktarux                                          *
  *                                                                         *
  *   This program is free software; you can redistribute it and/or modify  *
  *   it under the terms of the GNU General Public License as published by  *
@@ -32,6 +32,9 @@
 #ifdef USE_SRC
 #include <samplerate.h>
 #endif
+#ifdef USE_SPEEX
+#include <speex/speex_resampler.h>
+#endif
 
 #define M64P_PLUGIN_PROTOTYPES 1
 #include "m64p_types.h"
@@ -42,7 +45,6 @@
 #include "main.h"
 #include "volume.h"
 #include "osal_dynamiclib.h"
-#include "osal_preproc.h"
 
 /* Default start-time size of primary buffer (in equivalent output samples).
    This is the buffer where audio is loaded after it's extracted from n64's memory.
@@ -83,6 +85,16 @@ static int l_PluginInit = 0;
 static int l_PausedForSync = 1; /* Audio is started in paused state after SDL initialization */
 static m64p_handle l_ConfigAudio;
 
+enum resampler_type {
+	RESAMPLER_TRIVIAL,
+#ifdef USE_SRC
+	RESAMPLER_SRC,
+#endif
+#ifdef USE_SPEEX
+	RESAMPLER_SPEEX,
+#endif
+};
+
 /* Read header for type definition */
 static AUDIO_INFO AudioInfo;
 /* The hardware specifications we are using */
@@ -108,18 +120,21 @@ static unsigned int PrimaryBufferSize = PRIMARY_BUFFER_SIZE;
 static unsigned int PrimaryBufferTarget = PRIMARY_BUFFER_TARGET;
 // Size of Secondary audio buffer in output samples
 static unsigned int SecondaryBufferSize = SECONDARY_BUFFER_SIZE;
-// Resample or not
-static unsigned char Resample = 1;
+// Resample type
+static enum resampler_type Resample = RESAMPLER_TRIVIAL;
+// Resampler specific quality
+static int ResampleQuality = 3;
 // volume to scale the audio by, range of 0..100
+// if muted, this holds the volume when not muted
 static int VolPercent = 80;
 // how much percent to increment/decrement volume by
 static int VolDelta = 5;
 // the actual volume passed into SDL, range of 0..SDL_MIX_MAXVOLUME
 static int VolSDL = SDL_MIX_MAXVOLUME;
-// stores the previous volume when it is muted
-static int VolMutedSave = -1;
+// Muted or not
+static int VolIsMuted = 0;
 //which type of volume control to use
-static int VolumeControlType = VOLUME_TYPE_OSS;
+static int VolumeControlType = VOLUME_TYPE_SDL;
 
 static int OutputFreq;
 
@@ -259,8 +274,8 @@ EXPORT m64p_error CALL PluginStartup(m64p_dynlib_handle CoreLibHandle, void *Con
     ConfigSetDefaultInt(l_ConfigAudio, "PRIMARY_BUFFER_SIZE",   PRIMARY_BUFFER_SIZE,   "Size of primary buffer in output samples. This is where audio is loaded after it's extracted from n64's memory.");
     ConfigSetDefaultInt(l_ConfigAudio, "PRIMARY_BUFFER_TARGET", PRIMARY_BUFFER_TARGET, "Fullness level target for Primary audio buffer, in equivalent output samples");
     ConfigSetDefaultInt(l_ConfigAudio, "SECONDARY_BUFFER_SIZE", SECONDARY_BUFFER_SIZE, "Size of secondary buffer in output samples. This is SDL's hardware buffer.");
-    ConfigSetDefaultInt(l_ConfigAudio, "RESAMPLE",              1,                     "Audio resampling algorithm.  1 = unfiltered, 2 = SINC resampling (Best Quality, requires libsamplerate)");
-    ConfigSetDefaultInt(l_ConfigAudio, "VOLUME_CONTROL_TYPE",   VOLUME_TYPE_OSS,       "Volume control type: 1 = SDL (only affects Mupen64Plus output)  2 = OSS mixer (adjusts master PC volume)");
+    ConfigSetDefaultString(l_ConfigAudio, "RESAMPLE",              "trivial",                     "Audio resampling algorithm. src-sinc-best-quality, src-sinc-medium-quality, src-sinc-fastest, src-zero-order-hold, src-linear, speex-fixed-{10-0}, trivial");
+    ConfigSetDefaultInt(l_ConfigAudio, "VOLUME_CONTROL_TYPE",   VOLUME_TYPE_SDL,       "Volume control type: 1 = SDL (only affects Mupen64Plus output)  2 = OSS mixer (adjusts master PC volume)");
     ConfigSetDefaultInt(l_ConfigAudio, "VOLUME_ADJUST",         5,                     "Percentage change each time the volume is increased or decreased");
     ConfigSetDefaultInt(l_ConfigAudio, "VOLUME_DEFAULT",        80,                    "Default volume when a game is started.  Only used if VOLUME_CONTROL_TYPE is 1");
 
@@ -350,7 +365,7 @@ EXPORT void CALL AiLenChanged( void )
         return;
 
     LenReg = *AudioInfo.AI_LEN_REG;
-    p = (unsigned char*)(AudioInfo.RDRAM + (*AudioInfo.AI_DRAM_ADDR_REG & 0xFFFFFF));
+    p = AudioInfo.RDRAM + (*AudioInfo.AI_DRAM_ADDR_REG & 0xFFFFFF);
 
     if (buffer_pos + LenReg < primaryBufferBytes)
     {
@@ -448,14 +463,44 @@ static int error;
 static SRC_STATE *src_state;
 static SRC_DATA src_data;
 #endif
+#ifdef USE_SPEEX
+SpeexResamplerState* spx_state = NULL;
+static int error;
+#endif
 
 static int resample(unsigned char *input, int input_avail, int oldsamplerate, unsigned char *output, int output_needed, int newsamplerate)
 {
     int *psrc = (int*)input;
     int *pdest = (int*)output;
     int i = 0, j = 0;
+
+#ifdef USE_SPEEX
+    spx_uint32_t in_len, out_len;
+    if(Resample == RESAMPLER_SPEEX)
+    {
+        if(spx_state == NULL)
+        {
+            spx_state = speex_resampler_init(2, oldsamplerate, newsamplerate, ResampleQuality,  &error);
+            if(spx_state == NULL)
+            {
+                memset(output, 0, output_needed);
+                return 0;
+            }
+        }
+        speex_resampler_set_rate(spx_state, oldsamplerate, newsamplerate);
+        in_len = input_avail / 4;
+        out_len = output_needed / 4;
+
+        if ((error = speex_resampler_process_interleaved_int(spx_state, (const spx_int16_t *)input, &in_len, (spx_int16_t *)output, &out_len)))
+        {
+            memset(output, 0, output_needed);
+            return input_avail;  // number of bytes consumed
+        }
+        return in_len * 4;
+    }
+#endif
 #ifdef USE_SRC
-    if(Resample == 2)
+    if(Resample == RESAMPLER_SRC)
     {
         // the high quality resampler needs more input than the samplerate ratio would indicate to work properly
         if (input_avail > output_needed * 3 / 2)
@@ -476,7 +521,7 @@ static int resample(unsigned char *input, int input_avail, int oldsamplerate, un
         memset(_dest,0,_dest_len);
         if(src_state == NULL)
         {
-            src_state = src_new (SRC_SINC_BEST_QUALITY, 2, &error);
+            src_state = src_new (ResampleQuality, 2, &error);
             if(src_state == NULL)
             {
                 memset(output, 0, output_needed);
@@ -499,7 +544,7 @@ static int resample(unsigned char *input, int input_avail, int oldsamplerate, un
         return src_data.input_frames_used * 4;
     }
 #endif
-    // RESAMPLE == 1
+    // RESAMPLE == TRIVIAL
     if (newsamplerate >= oldsamplerate)
     {
         int sldf = oldsamplerate;
@@ -553,6 +598,7 @@ static void my_audio_callback(void *userdata, unsigned char *stream, int len)
 #endif
         {
             input_used = resample(primaryBuffer, buffer_pos, oldsamplerate, mixBuffer, len, newsamplerate);
+            memset(stream, 0, len);
             SDL_MixAudio(stream, mixBuffer, len, VolSDL);
         }
         memmove(primaryBuffer, &primaryBuffer[input_used], buffer_pos - input_used);
@@ -776,16 +822,118 @@ EXPORT void CALL SetSpeedFactor(int percentage)
 
 static void ReadConfig(void)
 {
+    const char *resampler_id;
+
     /* read the configuration values into our static variables */
     GameFreq = ConfigGetParamInt(l_ConfigAudio, "DEFAULT_FREQUENCY");
     SwapChannels = ConfigGetParamBool(l_ConfigAudio, "SWAP_CHANNELS");
     PrimaryBufferSize = ConfigGetParamInt(l_ConfigAudio, "PRIMARY_BUFFER_SIZE");
     PrimaryBufferTarget = ConfigGetParamInt(l_ConfigAudio, "PRIMARY_BUFFER_TARGET");
     SecondaryBufferSize = ConfigGetParamInt(l_ConfigAudio, "SECONDARY_BUFFER_SIZE");
-    Resample = ConfigGetParamInt(l_ConfigAudio, "RESAMPLE");
+    resampler_id = ConfigGetParamString(l_ConfigAudio, "RESAMPLE");
     VolumeControlType = ConfigGetParamInt(l_ConfigAudio, "VOLUME_CONTROL_TYPE");
     VolDelta = ConfigGetParamInt(l_ConfigAudio, "VOLUME_ADJUST");
     VolPercent = ConfigGetParamInt(l_ConfigAudio, "VOLUME_DEFAULT");
+
+    if (!resampler_id) {
+        Resample = RESAMPLER_TRIVIAL;
+	DebugMessage(M64MSG_WARNING, "Could not find RESAMPLE configuration; use trivial resampler");
+	return;
+    }
+    if (strcmp(resampler_id, "trivial") == 0) {
+        Resample = RESAMPLER_TRIVIAL;
+        return;
+    }
+#ifdef USE_SPEEX
+    if (strncmp(resampler_id, "speex-fixed-", strlen("speex-fixed-")) == 0) {
+        int i;
+        static const char *speex_quality[] = {
+            "speex-fixed-0",
+            "speex-fixed-1",
+            "speex-fixed-2",
+            "speex-fixed-3",
+            "speex-fixed-4",
+            "speex-fixed-5",
+            "speex-fixed-6",
+            "speex-fixed-7",
+            "speex-fixed-8",
+            "speex-fixed-9",
+            "speex-fixed-10",
+        };
+        Resample = RESAMPLER_SPEEX;
+        for (i = 0; i < sizeof(speex_quality) / sizeof(*speex_quality); i++) {
+            if (strcmp(speex_quality[i], resampler_id) == 0) {
+                ResampleQuality = i;
+                return;
+            }
+        }
+        DebugMessage(M64MSG_WARNING, "Unknown RESAMPLE configuration %s; use speex-fixed-4 resampler", resampler_id);
+        ResampleQuality = 4;
+        return;
+    }
+#endif
+#ifdef USE_SRC
+    if (strncmp(resampler_id, "src-", strlen("src-")) == 0) {
+        Resample = RESAMPLER_SRC;
+        if (strcmp(resampler_id, "src-sinc-best-quality") == 0) {
+            ResampleQuality = SRC_SINC_BEST_QUALITY;
+            return;
+        }
+        if (strcmp(resampler_id, "src-sinc-medium-quality") == 0) {
+            ResampleQuality = SRC_SINC_MEDIUM_QUALITY;
+            return;
+        }
+        if (strcmp(resampler_id, "src-sinc-fastest") == 0) {
+            ResampleQuality = SRC_SINC_FASTEST;
+            return;
+        }
+        if (strcmp(resampler_id, "src-zero-order-hold") == 0) {
+            ResampleQuality = SRC_ZERO_ORDER_HOLD;
+            return;
+        }
+        if (strcmp(resampler_id, "src-linear") == 0) {
+            ResampleQuality = SRC_LINEAR;
+            return;
+        }
+        DebugMessage(M64MSG_WARNING, "Unknown RESAMPLE configuration %s; use src-sinc-medium-quality resampler", resampler_id);
+        ResampleQuality = SRC_SINC_MEDIUM_QUALITY;
+        return;
+    }
+#endif
+    DebugMessage(M64MSG_WARNING, "Unknown RESAMPLE configuration %s; use trivial resampler", resampler_id);
+    Resample = RESAMPLER_TRIVIAL;
+}
+
+// Returns the most recent ummuted volume level.
+static int VolumeGetUnmutedLevel(void)
+{
+#if defined(HAS_OSS_SUPPORT)
+    // reload volume if we're using OSS
+    if (!VolIsMuted && VolumeControlType == VOLUME_TYPE_OSS)
+    {
+        return volGet();
+    }
+#endif
+
+    return VolPercent;
+}
+
+// Sets the volume level based on the contents of VolPercent and VolIsMuted
+static void VolumeCommit(void)
+{
+    int levelToCommit = VolIsMuted ? 0 : VolPercent;
+
+#if defined(HAS_OSS_SUPPORT)
+    if (VolumeControlType == VOLUME_TYPE_OSS)
+    {
+        //OSS mixer volume
+        volSet(levelToCommit);
+    }
+    else
+#endif
+    {
+        VolSDL = SDL_MIX_MAXVOLUME * levelToCommit / 100;
+    }
 }
 
 EXPORT void CALL VolumeMute(void)
@@ -793,40 +941,13 @@ EXPORT void CALL VolumeMute(void)
     if (!l_PluginInit)
         return;
 
-    if (VolMutedSave > -1)
-    {
-        //unmute
-        VolPercent = VolMutedSave;
-        VolMutedSave = -1;
-#if defined(HAS_OSS_SUPPORT)
-        if (VolumeControlType == VOLUME_TYPE_OSS)
-        {
-            //OSS mixer volume
-            volSet(VolPercent);
-        }
-        else
-#endif
-        {
-            VolSDL = SDL_MIX_MAXVOLUME * VolPercent / 100;
-        }
-    } 
-    else
-    {
-        //mute
-        VolMutedSave = VolPercent;
-        VolPercent = 0;
-#if defined(HAS_OSS_SUPPORT)
-        if (VolumeControlType == VOLUME_TYPE_OSS)
-        {
-            //OSS mixer volume
-            volSet(0);
-        }
-        else
-#endif
-        {
-            VolSDL = 0;
-        }
-    }
+    // Store the volume level in order to restore it later
+    if (!VolIsMuted)
+        VolPercent = VolumeGetUnmutedLevel();
+
+    // Toogle mute
+    VolIsMuted = !VolIsMuted;
+    VolumeCommit();
 }
 
 EXPORT void CALL VolumeUp(void)
@@ -834,34 +955,7 @@ EXPORT void CALL VolumeUp(void)
     if (!l_PluginInit)
         return;
 
-    //if muted, unmute first
-    if (VolMutedSave > -1)
-        VolumeMute();
-
-#if defined(HAS_OSS_SUPPORT)
-    // reload volume if we're using OSS
-    if (VolumeControlType == VOLUME_TYPE_OSS)
-    {
-        VolPercent = volGet();
-    }
-#endif
-
-    // adjust volume variable
-    VolPercent += VolDelta;
-    if (VolPercent > 100)
-        VolPercent = 100;
-
-#if defined(HAS_OSS_SUPPORT)
-    if (VolumeControlType == VOLUME_TYPE_OSS)
-    {
-        //OSS mixer volume
-        volSet(VolPercent);
-    }
-    else
-#endif
-    {
-        VolSDL = SDL_MIX_MAXVOLUME * VolPercent / 100;
-    }
+    VolumeSetLevel(VolumeGetUnmutedLevel() + VolDelta);
 }
 
 EXPORT void CALL VolumeDown(void)
@@ -869,39 +963,12 @@ EXPORT void CALL VolumeDown(void)
     if (!l_PluginInit)
         return;
 
-    //if muted, unmute first
-    if (VolMutedSave > -1)
-        VolumeMute();
-
-#if defined(HAS_OSS_SUPPORT)
-    // reload volume if we're using OSS
-    if (VolumeControlType == VOLUME_TYPE_OSS)
-    {
-        VolPercent = volGet();
-    }
-#endif
-
-    // adjust volume variable
-    VolPercent -= VolDelta;
-    if (VolPercent < 0)
-        VolPercent = 0;
-
-#if defined(HAS_OSS_SUPPORT)
-    if (VolumeControlType == VOLUME_TYPE_OSS)
-    {
-        //OSS mixer volume
-        volSet(VolPercent);
-    }
-    else
-#endif
-    {
-        VolSDL = SDL_MIX_MAXVOLUME * VolPercent / 100;
-    }
+    VolumeSetLevel(VolumeGetUnmutedLevel() - VolDelta);
 }
 
 EXPORT int CALL VolumeGetLevel(void)
 {
-    return VolPercent;
+    return VolIsMuted ? 0 : VolumeGetUnmutedLevel();
 }
 
 EXPORT void CALL VolumeSetLevel(int level)
@@ -909,6 +976,9 @@ EXPORT void CALL VolumeSetLevel(int level)
     if (!l_PluginInit)
         return;
 
+    //if muted, unmute first
+    VolIsMuted = 0;
+
     // adjust volume 
     VolPercent = level;
     if (VolPercent < 0)
@@ -916,24 +986,14 @@ EXPORT void CALL VolumeSetLevel(int level)
     else if (VolPercent > 100)
         VolPercent = 100;
 
-#if defined(HAS_OSS_SUPPORT)
-    if (VolumeControlType == VOLUME_TYPE_OSS)
-    {
-        //OSS mixer volume
-        volSet(VolPercent);
-    }
-    else
-#endif
-    {
-        VolSDL = SDL_MIX_MAXVOLUME * VolPercent / 100;
-    }
+    VolumeCommit();
 }
 
-static char VolumeString[32];
-
 EXPORT const char * CALL VolumeGetString(void)
 {
-    if (VolMutedSave > -1)
+    static char VolumeString[32];
+
+    if (VolIsMuted)
     {
         strcpy(VolumeString, "Mute");
     }
